#include "config.h"
#include "mac.h"

#include <mac/APETag.h>
#include <mac/APEInfo.h>
#include <mac/CharacterHelper.h>

#include <audacious/util.h>
#include <audacious/main.h>

#include <glib.h>

const gchar *fmts[] = { "mac", "ape", "apl", NULL };

InputPlugin mac_plugin_info =
{
    NULL,
    NULL,
    (gchar *)"Monkeys Audio Codec Player " VERSION,
    NULL,
    NULL,
    mac_about,
    NULL,
    FALSE,
    NULL,
    NULL,
    mac_play_file,
    mac_stop,
    mac_pause,
    mac_seek, 
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    mac_file_info_box,
    NULL,
    mac_get_song_tuple,
    NULL,
    NULL,
    mac_is_our_file_from_vfs,
    (gchar **)fmts,
    mac_mseek,
};

InputPlugin *mac_plugin_infolist[] = { &mac_plugin_info, NULL };

SIMPLE_INPUT_PLUGIN(mac_plugin_info, mac_plugin_infolist)

static PlayerInfo *mac_info;

#ifdef __cplusplus
extern "C"{
#endif

static char *get_tag_info(CAPETag *tag, wchar_t *fieldname)
{
    if (!tag)
    {
	return NULL;
    }


    CAPETagField *pTagField = tag->GetTagField(fieldname);

    if (pTagField == NULL)
    {
	return g_strdup("");
    }

    const char *fieldValue;
    char *value;

    fieldValue = pTagField->GetFieldValue();
    if (tag->GetAPETagVersion() == CURRENT_APE_TAG_VERSION)
    {
	value = GetANSIFromUTF8((unsigned char *)fieldValue);
    }
    else
    {
	value = (char *)g_malloc0(255);
	strncpy(value, fieldValue, 255);
    }

    char *ret = g_strdup(value);
    free(value);
    
    return ret;
}

#define MAC_aud_tuple_associate_string(apetagfield, tuplename) \
    scratch = get_tag_info(tag, apetagfield); \
    if(scratch) { \
        if (*scratch) \
            aud_tuple_associate_string(input, tuplename, NULL, scratch); \
        g_free(scratch); scratch = NULL; \
    };

#define MAC_aud_tuple_associate_string_custom(apetagfield, tuplename) \
    scratch = get_tag_info(tag, apetagfield); \
    if(scratch) { \
        if (*scratch) \
            aud_tuple_associate_string(input, -1, tuplename, scratch); \
        g_free(scratch); scratch = NULL; \
    };

Tuple *get_song_tuple_from_iapedec(gchar *filename, IAPEDecompress *decompress)
{
    char *scratch = NULL;

    CAPETag *tag = (CAPETag *)decompress->GetInfo(APE_INFO_TAG);
    Tuple *input = aud_tuple_new_from_filename(filename);

    if (tag && (tag->GetHasID3Tag() || tag->GetHasAPETag()))
    {
        MAC_aud_tuple_associate_string(APE_TAG_FIELD_ALBUM, FIELD_ALBUM);
        MAC_aud_tuple_associate_string(APE_TAG_FIELD_ARTIST, FIELD_ARTIST);
        MAC_aud_tuple_associate_string(APE_TAG_FIELD_COMMENT, FIELD_COMMENT);
        MAC_aud_tuple_associate_string(APE_TAG_FIELD_GENRE, FIELD_GENRE);
        MAC_aud_tuple_associate_string(APE_TAG_FIELD_TITLE, FIELD_TITLE);
        MAC_aud_tuple_associate_string_custom(APE_TAG_FIELD_COMPOSER, "composer");
        

        aud_tuple_associate_int(input, FIELD_TRACK_NUMBER, NULL, atoi(get_tag_info(tag, APE_TAG_FIELD_TRACK)) );
        aud_tuple_associate_int(input, FIELD_YEAR, NULL, atoi(get_tag_info(tag, APE_TAG_FIELD_YEAR)) );
    }/* else {
        scratch = g_path_get_basename(filename);
        aud_tuple_associate_string(input, "title", scratch);
        g_free(scratch); scratch = NULL;
    } */
    /* Should be handled by aud_tuple_formatter_make_title_string now.*/

    aud_tuple_associate_string(input, FIELD_QUALITY, NULL, (gchar *)"lossy");
    aud_tuple_associate_string(input, FIELD_CODEC, NULL, (gchar *)"Monkey's Audio Codec (MAC)");

    aud_tuple_associate_int(input, FIELD_LENGTH, NULL, decompress->GetInfo(APE_DECOMPRESS_LENGTH_MS) );

    return input;
}

static char *mac_format_title_string(char *filename, IAPEDecompress *iapedec)
{
    char *title = NULL; 
    Tuple *input = get_song_tuple_from_iapedec(filename, iapedec);

    title = aud_tuple_formatter_make_title_string(input, aud_get_gentitle_format());
    
    aud_tuple_free(input);

    return title;
}

static int decompress_init(InputPlayback *playback)
{
    int nRetVal = 0;
    char *uriname = playback->filename;
    char *filename;

    if (uriname == NULL || (strncmp(uriname, "file:/", 6) != 0) )
        return -1;

    mac_info = (PlayerInfo *)g_malloc0(sizeof(PlayerInfo));

    playback->playing = false;
    playback->eof = false;
    mac_info->seek_to = -1;

    IAPEDecompress *decompress = NULL;
    
    wchar_t *pUTF16;

    filename = g_filename_from_uri(uriname, NULL, NULL);
    pUTF16= GetUTF16FromANSI(filename);
     
    decompress = CreateIAPEDecompress(pUTF16, &nRetVal);
    free(pUTF16);

    if (!decompress || nRetVal != ERROR_SUCCESS)
    {
	return -1;
    }    

    mac_info->pAPEDecompress = decompress;

    mac_info->title = mac_format_title_string(filename, decompress);

    mac_info->sample_rate = decompress->GetInfo(APE_INFO_SAMPLE_RATE);
    mac_info->bits_per_sample = decompress->GetInfo(APE_INFO_BITS_PER_SAMPLE);
    mac_info->channels = decompress->GetInfo(APE_INFO_CHANNELS);
    mac_info->length_in_ms = decompress->GetInfo(APE_DECOMPRESS_LENGTH_MS);
    mac_info->block_align = decompress->GetInfo(APE_INFO_BLOCK_ALIGN);

    mac_info->sample_format = (mac_info->bits_per_sample == 16) ? FMT_S16_LE : FMT_S8;
    mac_info->seek_to = -1;
    playback->eof = false;

/*
    printf("title = %s\n", mac_info->title);
    printf("sample = %d\n", mac_info->sample_rate);
    printf("bits per sampler = %d\n", mac_info->bits_per_sample);
    printf("channels = %d\n", mac_info->channels);
    printf("length = %d\n", mac_info->length_in_ms);
    printf("sample format = %s\n", (mac_info->sample_format == FMT_S16_LE) ? "FMT_S16_LE" : "FMT_S8");
*/
    g_free(filename);
    return 0;
}

#define SAMPLES_PER_READ 512

static void *decode_loop(void *arg)
{
    InputPlayback *playback = (InputPlayback *) arg;
    
    char data[9216];

    int blocks_to_read, actual_read;
    int nRetVal;
    int bytes, actual_bytes;

    blocks_to_read = SAMPLES_PER_READ;
    bytes = blocks_to_read * mac_info->block_align;

    while (playback->playing)
    {
	if (mac_info->seek_to != -1)
	{
        int to_block = (int)((gint64)mac_info->seek_to * (gint64)mac_info->sample_rate / 1000L);
	    mac_info->pAPEDecompress->Seek(to_block);

	    playback->output->flush(mac_info->seek_to);
	    mac_info->seek_to = -1;
	}

	if (!playback->eof)
	{
	    actual_read = blocks_to_read;
	    actual_bytes = bytes;

	    nRetVal = mac_info->pAPEDecompress->GetData(data, blocks_to_read, &actual_read);

	    if (actual_read)
	    {
		if (actual_read < blocks_to_read)
		{
		    actual_bytes = actual_read * mac_info->block_align;
		}
		bytes = blocks_to_read * (mac_info->bits_per_sample / 8) * mac_info->channels;

		if (playback->playing && mac_info->seek_to == -1)
		{
		    playback->pass_audio(playback, (mac_info->sample_format == FMT_S16_LE) ? FMT_S16_LE : FMT_S8,
			mac_info->channels, bytes, data, &playback->playing);
		}

	    }
	    else
		playback->eof = true;
	}
	else
	{
	    playback->output->buffer_free ();
	    playback->output->buffer_free ();
	    while (playback->output->buffer_playing())
	        g_usleep(10000);
	    playback->playing = 0;
	}
    }
    return NULL;
}

/***********************************************
 *                                             *
 *             Interface Functions             *
 *                                             *
 ***********************************************/

void mac_about()
{
    static GtkWidget *aboutbox = NULL;

    if (aboutbox != NULL)
	return;
	
    aboutbox = audacious_info_dialog(
	_("About Monkey's Audio Codec plugin"),
	_("MAC decoding engine by Matthew T. Ashland <email@monkeysaudio.com>\n\nPlugin by SuperMMX <SuperMMX@163.com>\n\nAudacious port by Christian Birchinger <joker@netswarm.net>"),
	_("OK"), FALSE, NULL, NULL);

    g_signal_connect(GTK_OBJECT(aboutbox), "destroy",
		     GTK_SIGNAL_FUNC(gtk_widget_destroyed), &aboutbox);
}

int mac_is_our_file_from_vfs(gchar *filename, VFSFile *file)
{
    gchar magic[4];
    const gchar *ext;
    unsigned char filehead[] = { 0x4D, 0x41, 0x43, 0x20 }; // "MAC " - Monkey's Audio
    
    aud_vfs_fread(magic, 1, 4, file);
    if (!memcmp(magic, filehead, 4))
        return TRUE;
    
    ext = strrchr(filename, '.');
    if (ext) {
        if (!strcasecmp(ext, ".mac")) return TRUE;
        if (!strcasecmp(ext, ".ape")) return TRUE;
        if (!strcasecmp(ext, ".apl")) return TRUE;
    }
    
    return FALSE;
}

void mac_play_file(InputPlayback *playback)
{
    guint32 bitrate = 0;

    if (decompress_init(playback) < 0)
    {
	return;
    }

    playback->playing = true;

    if(playback->output->open_audio(mac_info->sample_format, mac_info->sample_rate, mac_info->channels) == 0) 
    {
	return;
    }

    bitrate = mac_info->pAPEDecompress->GetInfo(APE_DECOMPRESS_AVERAGE_BITRATE);
    playback->set_params(playback, mac_info->title, mac_info->length_in_ms, 
			     bitrate * 1000,
			     mac_info->sample_rate, mac_info->channels);

    mac_info->decoder_thread = g_thread_self();
    playback->set_pb_ready(playback);
    decode_loop(playback);
}

void mac_stop(InputPlayback *playback)
{
    if(playback->playing) 
    {
	playback->playing = false;
	g_thread_join(mac_info->decoder_thread);
	playback->output->close_audio();
	
	if (mac_info->pAPEDecompress)
	{
	    delete mac_info->pAPEDecompress;
	    mac_info->pAPEDecompress = NULL;
	}
	g_free(mac_info);
	mac_info = NULL;
    }

}

void mac_pause(InputPlayback *playback, short paused)
{
    playback->output->pause(paused);
}

void mac_mseek(InputPlayback *playback, gulong millisecond)
{

    mac_info->seek_to = (gint)millisecond;
    playback->eof = false;

    while (mac_info && mac_info->seek_to != -1)
    {
	g_usleep(10000);
    }

}

void mac_seek(InputPlayback *playback, int time)
{
    mac_mseek(playback, time * 1000);
}

Tuple *mac_get_song_tuple(gchar *uriname)
{
    int nRetVal = 0;
    char *filename;
    Tuple *input;
    
    if (uriname == NULL || (strncmp(uriname, "file:/", 6) != 0) )
	return NULL;

    filename = g_filename_from_uri(uriname, NULL, NULL);

    wchar_t *pUTF16Name = GetUTF16FromANSI(filename);
    IAPEDecompress *pDec = CreateIAPEDecompress(pUTF16Name, &nRetVal);

    if (nRetVal != ERROR_SUCCESS) {
        g_free(filename);
        return NULL;
    }

    input = get_song_tuple_from_iapedec(filename, pDec);

    g_free(filename);
    delete pDec;

    return input;
}

#ifdef __cplusplus
}
#endif
